unit accesnt;

interface

uses
  Windows, SysUtils, Classes, WinSvc;

const
    METHOD_BUFFERED = 0;
    FILE_READ_ACCESS = 1;
    FILE_WRITE_ACCESS = 2;
    GPD_TYPE = $9C40;
    READ_BYTE = $900;
    READ_WORD = $901;
    READ_DWORD = $902;
    WRITE_BYTE = $910;
    WRITE_WORD = $911;
    WRITE_DWORD = $912;
    ReadIOByte   = DWORD((GPD_TYPE shl 16) or (FILE_READ_ACCESS  shl 14) or (READ_BYTE shl 2) or METHOD_BUFFERED);
    WriteIOByte  = DWORD((GPD_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or (WRITE_BYTE shl 2) or METHOD_BUFFERED);
    ReadIOWord   = DWORD((GPD_TYPE shl 16) or (FILE_READ_ACCESS  shl 14) or (READ_WORD shl 2) or METHOD_BUFFERED);
    WriteIOWord  = DWORD((GPD_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or (WRITE_WORD shl 2) or METHOD_BUFFERED);
    ReadIODWord  = DWORD((GPD_TYPE shl 16) or (FILE_READ_ACCESS  shl 14) or (READ_DWORD shl 2) or METHOD_BUFFERED);
    WriteIODWord = DWORD((GPD_TYPE shl 16) or (FILE_WRITE_ACCESS shl 14) or (WRITE_DWORD shl 2) or METHOD_BUFFERED);
    IOCTL_INSB   = DWORD((GPD_TYPE shl 16) or (FILE_READ_ACCESS  shl 14) or ($920 shl 2) or METHOD_BUFFERED);
    IOCTL_INSW   = DWORD((GPD_TYPE shl 16) or (FILE_READ_ACCESS  shl 14) or ($921 shl 2) or METHOD_BUFFERED);
    IOCTL_INSD   = DWORD((GPD_TYPE shl 16) or (FILE_READ_ACCESS  shl 14) or ($922 shl 2) or METHOD_BUFFERED);

    DriverName : PChar = 'ACCESNT';
    SERVICE_AUTO_START = $02;
    SERVICE_DEMAND_START = $03;
    SERVICE_ERROR_NORMAL = $01;
    SERVICE_KERNEL_DRIVER = $01;

var
    Args : PChar = NIL;

type TDataStruct = record
    Address: DWORD;
    Value: DWORD;
    end;

procedure InitACCESNT;
procedure CloseACCESNT;
function GotACCESNT: Boolean;
function NT_InPortB(Address: DWORD): WORD;
function NT_InPort(Address: DWORD): WORD;
function NT_InPortL(Address: DWORD): DWORD;
function NT_OutPortB(Address: DWORD; Value: BYTE): WORD;
function NT_OutPort(Address: DWORD; Value: WORD): WORD;
function NT_OutPortL(Address: DWORD; Value: DWORD): WORD;

function NT_INSB(Port, Count: LongWord; var pBuffer:     Byte): LongWord; cdecl;
function NT_INSW(Port, Count: LongWord; var pBuffer:     Word): LongWord; cdecl;
function NT_INSD(Port, Count: LongWord; var pBuffer: LongWord): LongWord; cdecl;

implementation

var
    hDevice: LongWord = INVALID_HANDLE_VALUE;
    ManagerHandle, ServiceHandle : SC_HANDLE;
    Ptp: TTokenPrivileges;

const
    ServiceExe: PChar = '%SystemRoot%\SYSTEM32\DRIVERS\accesnt.sys';

procedure BypassRights;
var
    tp: TTokenPrivileges;
    luid: TLargeInteger;
    hToken: THandle;
    cbRet: LongWord;
begin
    LookupPrivilegeValue(nil, 'SeLoadDriverPrivilege', luid);

    tp.PrivilegeCount := 1;
    tp.Privileges[0].Luid := luid;
    tp.Privileges[0].Attributes := SE_PRIVILEGE_ENABLED;

    OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY, hToken);
    AdjustTokenPrivileges(hToken,
                          FALSE,
                          tp,
                          sizeof(tp),
                          Ptp,
                          cbRet);
end;

procedure RestoreRights;
var
    tp: TTokenPrivileges;
    luid: TLargeInteger;
    hToken: THandle;
    cbRet: LongWord;
begin
    LookupPrivilegeValue(nil, 'SeLoadDriverPrivilege', luid);
    OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, hToken);
    AdjustTokenPrivileges(hToken,
                          FALSE,
                          Ptp,
                          sizeof(Ptp),
                          tp,
                          cbRet);
end;

procedure InitACCESNT;
begin
    if hDevice <> INVALID_HANDLE_VALUE then Exit;

    hDevice := CreateFile('\\.\ACCESNT',
                          GENERIC_WRITE or GENERIC_READ,
                          FILE_SHARE_READ or FILE_SHARE_WRITE,
                          nil,
                          OPEN_EXISTING,
                          0,
                          0
    );

    if hDevice <> INVALID_HANDLE_VALUE then Exit;

    BypassRights;
  // open manager
    ManagerHandle := OpenSCManager (NIL, NIL, SC_MANAGER_ALL_ACCESS);
  // create service
    ServiceHandle := CreateService (ManagerHandle,        // SCManager database
                                    DriverName,         // name of service
                                    DriverName,         // name to display
                                    SERVICE_ALL_ACCESS,    // desired access
                                    SERVICE_KERNEL_DRIVER, // service type
                                    SERVICE_AUTO_START,  // start type
                                    SERVICE_ERROR_NORMAL,  // error control type
                                    ServiceExe,          // service's binary
                                    NIL,               // no load ordering group
                                    NIL,               // no tag identifier
                                    NIL,               // no dependencies
                                    NIL,               // LocalSystem account
                                    NIL                // no password
    );

  // open if already created
    if (ServiceHandle = 0) then
        ServiceHandle := OpenService(ManagerHandle, DriverName, SERVICE_ALL_ACCESS);

    //DeleteService (ServiceHandle);
    StartService (ServiceHandle, 0, Args);

    CloseServiceHandle (ServiceHandle);
    CloseServiceHandle (ManagerHandle);

    hDevice := CreateFile('\\.\ACCESNT',
                          GENERIC_WRITE or GENERIC_READ,
                          FILE_SHARE_READ or FILE_SHARE_WRITE,
                          nil,
                          OPEN_EXISTING,
                          0,
                          0
    );
end;

procedure CloseACCESNT;
var
   ServiceStatus : TServiceStatus;
begin
    if hDevice <> 0 then
        CloseHandle(hDevice);
Exit; // if autostart
   ManagerHandle := OpenSCManager (NIL, NIL, SC_MANAGER_ALL_ACCESS);
   ServiceHandle := OpenService(ManagerHandle, DriverName, SERVICE_ALL_ACCESS);
   ControlService(ServiceHandle, SERVICE_CONTROL_STOP, ServiceStatus);
   CloseServiceHandle(ManagerHandle);
   CloseServiceHandle(ServiceHandle);
//   RestoreRights;
end;

function GotACCESNT: Boolean;
begin
    Result := hDevice <> INVALID_HANDLE_VALUE;
end;

function NT_InPortB(Address: DWORD): WORD;
var
    DataBuffer: DWORD;
    cbRet: DWORD;
    Success: Boolean;
begin
    Success := DeviceIoControl(hDevice,
                            ReadIOByte,
                            @Address,
                            sizeof(Address),
                            @DataBuffer,
                            sizeof(DataBuffer),
                            cbRet,
                            nil);
    if Success then
        Result := DataBuffer and $FF
    else
        Result := $AA55;
end;

function NT_InPort(Address: DWORD): WORD;
var
    DataBuffer: DWORD;
    cbRet: DWORD;
    Success: Boolean;
begin
    Success := DeviceIoControl(hDevice,
                            ReadIOWord,
                            @Address,
                            sizeof(Address),
                            @DataBuffer,
                            sizeof(DataBuffer),
                            cbRet,
                            nil);
    if Success then
        Result := DataBuffer and $FFFF
    else
        Result := $AA55;
end;

function NT_InPortL(Address: DWORD): DWORD;
var
    DataBuffer: DWORD;
    cbRet: DWORD;
    Success: Boolean;
begin
    Success := DeviceIoControl(hDevice,
                            ReadIODWord,
                            @Address,
                            sizeof(Address),
                            @DataBuffer,
                            sizeof(DataBuffer),
                            cbRet,
                            nil);
    if Success then
        Result := DataBuffer
    else
        Result := $AA55;
end;

function NT_OutPortB(Address: DWORD; Value: BYTE): WORD;
var
    cbRet: DWORD;
    DataStruct: TDataStruct;
    Success: Boolean;
begin
    DataStruct.Address := Address;
    DataStruct.Value := Value;

    Success := DeviceIoControl(hDevice,
                            WriteIOByte,
                            @DataStruct,
                            sizeof(DataStruct),
                            nil,
                            0,
                            cbRet,
                            nil);
    if Success then
        Result := 0
    else
        Result := $AA55;
end;

function NT_OutPort(Address: DWORD; Value: WORD): WORD;
var
    cbRet: DWORD;
    DataStruct: TDataStruct;
    Success: Boolean;
begin
    DataStruct.Address := Address;
    DataStruct.Value := Value;

    Success := DeviceIoControl(hDevice,
                            WriteIOWord,
                            @DataStruct,
                            sizeof(DataStruct),
                            nil,
                            0,
                            cbRet,
                            nil);
    if Success then
        Result := 0
    else
        Result := $AA55;
end;

function NT_OutPortL(Address: DWORD; Value: DWORD): WORD;
var
    cbRet: DWORD;
    DataStruct: TDataStruct;
    Success: Boolean;
begin
    DataStruct.Address := Address;
    DataStruct.Value := Value;

    Success := DeviceIoControl(hDevice,
                            WriteIODWord,
                            @DataStruct,
                            sizeof(DataStruct),
                            nil,
                            0,
                            cbRet,
                            nil);
    if Success then
        Result := 0
    else
        Result := $AA55;
end;

function GetMappedFailure: LongWord;
begin
  Result := GetLastError;
  if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
end;

function NT_INSB(Port, Count: LongWord; var pBuffer:     Byte): LongWord;
var
  cbRet: DWORD;

begin
  if DeviceIoControl(hDevice, IOCTL_INSB,
                     @Port, sizeof(Port),
                     @pBuffer, Count * sizeof(pBuffer),
                     cbRet, nil
  ) then
    Result := ERROR_SUCCESS
  else
    Result := GetMappedFailure
  ;
end;

function NT_INSW(Port, Count: LongWord; var pBuffer:     Word): LongWord;
var
  cbRet: DWORD;

begin
  if DeviceIoControl(hDevice, IOCTL_INSW,
                     @Port, sizeof(Port),
                     @pBuffer, Count * sizeof(pBuffer),
                     cbRet, nil
  ) then
    Result := ERROR_SUCCESS
  else
    Result := GetMappedFailure
  ;
end;

function NT_INSD(Port, Count: LongWord; var pBuffer: LongWord): LongWord;
var
  cbRet: DWORD;

begin
  if DeviceIoControl(hDevice, IOCTL_INSD,
                     @Port, sizeof(Port),
                     @pBuffer, Count * sizeof(pBuffer),
                     cbRet, nil
  ) then
    Result := ERROR_SUCCESS
  else
    Result := GetMappedFailure
  ;
end;

end.
